use std::{collections::HashSet, sync::Arc};

use anyhow::Result;
use async_recursion::async_recursion;
use dbus::nonblock::SyncConnection;
use futures::{stream, StreamExt};
use serde_derive::{Deserialize, Serialize};

use crate::{
    dbus_wrappers::{
        active_connection::ActiveConnectionWrapper, connectivity_state::ConnectivityState,
        device::DeviceWrapper, device_state::DeviceState, manager::ManagerWrapper, state::State,
    },
    identifier::{ActiveConnectionIdentifier, DeviceIdentifier},
};

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum Condition {
    // Always passes.
    AlwaysTrue,
    // Always fails.
    AlwaysFalse,
    // Passes if the last connectivity check is any of the specificed states.
    ConnectivityStateIsAnyOf {
        states: HashSet<ConnectivityState>,
    },
    // Passes if the device has any active connections.
    DeviceIsConnected {
        device_identifier: DeviceIdentifier,
    },
    // Fails if the device has any active connections.
    DeviceIsNotConnected {
        device_identifier: DeviceIdentifier,
    },
    // Passes if the device has any of the listed states.
    DeviceStateIsAnyOf {
        device_identifier: DeviceIdentifier,
        states: HashSet<DeviceState>,
    },
    // Passes if the device is connected to any of the listed connections.
    DeviceIsConnectedToOneOf {
        device_identifier: DeviceIdentifier,
        connections: Vec<ActiveConnectionIdentifier>,
    },
    // Passes if the device is not connected to the listed connections.
    DeviceIsNotConnectedToOneOf {
        device_identifier: DeviceIdentifier,
        connections: Vec<ActiveConnectionIdentifier>,
    },
    // TODO: Passes if the device is connected to one of the listed SSIDs.
    // DeviceIsConnectedToOneOfSSIDs {
    //     device_identifier: DeviceIdentifier,
    //     ssids: Vec<String>,
    // },
    // TODO: Passes if the device is not connected to the listed SSIDs.
    // DeviceIsNotConnectedToOneOfSSIDs {
    //     device_identifier: DeviceIdentifier,
    //     ssids: Vec<String>,
    // },
    // Passes if the current NetworkManager state is one of the specified
    // states.
    StateIsAnyOf {
        states: HashSet<State>,
    },
    Not {
        condition: Box<Condition>,
    },
}

impl Condition {
    #[async_recursion]
    pub async fn evaluate(&self, conn: &Arc<SyncConnection>) -> Result<bool> {
        match self {
            Condition::AlwaysTrue => Ok(true),
            Condition::AlwaysFalse => Ok(false),
            Condition::ConnectivityStateIsAnyOf { states } => {
                log::debug!("Evaluating ConnectivityStateIsAnyOf condition");

                let manager = ManagerWrapper::from_connection(conn).await;

                Ok(states.contains(&manager.get_last_connectivity_state().await?))
            }
            Condition::DeviceIsConnected {
                device_identifier: device_id,
            } => {
                log::debug!("Evaluating DeviceIsConnected condition for {:?}", device_id);

                match device_id.to_device(conn).await? {
                    Some(device) => {
                        device_matches_states(
                            &device,
                            vec![DeviceState::Activated].into_iter().collect(),
                        )
                        .await
                    }
                    None => Ok(false),
                }
            }
            Condition::DeviceIsNotConnected {
                device_identifier: device_id,
            } => {
                log::debug!(
                    "Evaluating DeviceIsNotConnected condition for {:?}",
                    device_id
                );

                match device_id.to_device(conn).await? {
                    Some(device) => {
                        device_matches_states(
                            &device,
                            vec![DeviceState::Disconnected, DeviceState::Unavailable]
                                .into_iter()
                                .collect(),
                        )
                        .await
                    }
                    None => Ok(true),
                }
            }
            Condition::DeviceStateIsAnyOf {
                device_identifier: device_id,
                states,
            } => {
                log::debug!(
                    "Evaluating DeviceStateIsAnyOf condition for {:?}",
                    device_id
                );

                match device_id.to_device(conn).await? {
                    Some(device) => device_matches_states(&device, states.clone()).await,
                    None => Ok(false),
                }
            }
            Condition::DeviceIsConnectedToOneOf {
                device_identifier: device_id,
                connections,
            } => {
                log::debug!(
                    "Evaluating DeviceIsConnectedToOneOf condition for {:?}",
                    device_id
                );

                device_is_connected_to_one_of(conn, device_id, connections).await
            }
            Condition::DeviceIsNotConnectedToOneOf {
                device_identifier: device_id,
                connections,
            } => {
                log::debug!(
                    "Evaluating DeviceIsConnectedToOneOf condition for {:?}",
                    device_id
                );

                Ok(!device_is_connected_to_one_of(conn, device_id, connections).await?)
            }
            Condition::StateIsAnyOf { states } => {
                log::debug!("Evaluating StateIsAnyOf condition");

                let manager = ManagerWrapper::from_connection(conn).await;

                Ok(states.contains(&manager.get_state().await?))
            }
            Condition::Not { condition } => {
                let result = condition.evaluate(conn).await?;

                Ok(!result)
            }
        }
    }
}

async fn device_matches_states(
    device: &DeviceWrapper<'_>,
    states: HashSet<DeviceState>,
) -> Result<bool> {
    match device.get_state().await? {
        Some(state) => Ok(states.contains(&state)),
        _ => Ok(false),
    }
}

async fn device_is_connected_to_one_of(
    conn: &Arc<SyncConnection>,
    device_identifier: &DeviceIdentifier,
    connections: &[ActiveConnectionIdentifier],
) -> Result<bool> {
    match device_identifier.to_device(conn).await? {
        Some(device) => {
            let active_connections: Vec<ActiveConnectionWrapper<'_>> = stream::iter(connections)
                .filter_map(|connection_identifier| async move {
                    match connection_identifier.to_active_connection(conn).await {
                        Ok(active_connection) => active_connection,
                        Err(err) => {
                            log::error!(
                                "Got error retrieving active connection for {:?}: {:?}",
                                connection_identifier,
                                err
                            );

                            None
                        }
                    }
                })
                .collect()
                .await;

            let has_any_matches = stream::iter(active_connections).any(|active_connection| {
                let device_path = device.get_path();

                async move {
                    let device_paths = active_connection.get_device_paths().await;

                    match device_paths {
                        Ok(device_paths) => device_paths.iter().any(|x| x.eq(&device_path)),
                        Err(err) => {
                            log::error!(
                                "Unable to get device paths for active connection {:?}: {:?}",
                                active_connection.get_path(),
                                err
                            );

                            false
                        }
                    }
                }
            });

            Ok(has_any_matches.await)
        }
        None => Ok(false),
    }
}
